import sqlite3
import json
from pathlib import Path

current_dir = Path(__file__).parent
input_path = current_dir / "pypitzer_parameter.json"

data = json.load(open(input_path))

output_path = current_dir / "pypitzer.sqlite"

conn = sqlite3.connect(output_path)
cur = conn.cursor()

def coeffs(entry, key):
    d = entry.get(key, {})
    return ",".join(str(d.get(f"a{i}", 0)) for i in range(6))

cur.executescript("""
CREATE TABLE IF NOT EXISTS binary_params (
    pair TEXT PRIMARY KEY,
    ref TEXT,
    T_min REAL,
    T_max REAL,
    eq_num INTEGER,
    b0 TEXT,
    b1 TEXT,
    b2 TEXT,
    c_phi TEXT,
    lambda TEXT,
    theta TEXT
);

CREATE TABLE IF NOT EXISTS ternary_params (
    triplet TEXT PRIMARY KEY,
    ref TEXT,
    T_min REAL,
    T_max REAL,
    eq_num INTEGER,
    psi TEXT,
    zeta TEXT
);
""")

# Insert binary
for pair, entry in data["binary"].items():
    cur.execute("""
        INSERT OR REPLACE INTO binary_params
        (pair, ref, T_min, T_max, eq_num, b0, b1, b2, c_phi, lambda, theta)
        VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
    """, (
        pair,
        entry["ref"],
        entry["T_range"][0],
        entry["T_range"][1],
        entry["eq_num"],
        coeffs(entry, "b0"),
        coeffs(entry, "b1"),
        coeffs(entry, "b2"),
        coeffs(entry, "c_phi"),
        coeffs(entry, "lambda"),
        coeffs(entry, "theta"),
    ))

# Insert ternary
for triplet, entry in data["ternary"].items():
    cur.execute("""
        INSERT OR REPLACE INTO ternary_params
        (triplet, ref, T_min, T_max, eq_num, psi, zeta)
        VALUES (?, ?, ?, ?, ?, ?, ?)
    """, (
        triplet,
        entry["ref"],
        entry["T_range"][0],
        entry["T_range"][1],
        entry["eq_num"],
        coeffs(entry, "psi"),
        coeffs(entry, "zeta"),
    ))

conn.commit()
conn.close()



# -*- coding: utf-8 -*-
# Author: Yiping Liu
# Description:
# Version: 1.0
# Last Modified: Nov 03, 2025

import time

import numpy as np
from scipy.optimize import minimize

from pypitzer.database.db_query import PypitzerDB
from pypitzer.public.low_level import type_of_species, get_charge_number
from pypitzer.public.icemelting import (clegg_and_brimblecombe, spencer, monnin)

import pypitzer.Pitzer.methods as pm

db = PypitzerDB()


class FluidPitzer:
    def __init__(self, x0, species, equilibrium, t=25.0, neutral=False, ):
        """
        Initiate the solution.
        :param x0: initiate values of x(x1, x2), namely value of mNa and mCl, [mol/kg].
        :param species: a dict containing aqueous spcies and their relative concentrations to Na, [-].
        :param t: melting temperature, [°C].
        :param equilibrium: str, a chemical expression indicating the solubility equilibrium, e.g., "NaCl(s) = Na+(aq) + Cl-(aq)".
        """
        self.x0 = x0
        self.t = t
        self.T = t + 273.16
        self.input_species = species
        self.equilibrium = equilibrium
        self.neutral = neutral  # to determine if a neutral species is considered in this fluid.
        self.solid_disso = db.get_reaction(equilibrium, self.t)

        self._species = None
        self._component_groups = None
        self._param_db = None

    @property
    def species(self):
        if self._species is None :
            species = self.input_species
            species['Cl-'] = 0 # add Cl into the dict and set a default value of 0.
            self._species = species
        return  self._species
    
    @property
    def component_groups(self):
        if self._component_groups is None:
            components = tuple(self.species.keys())
            self._component_groups = pm.group_components(components)
        return self._component_groups
    
    @staticmethod
    def order_str(tuple_pair):
        """
        Order elements of a tuple alphabetically and join them as a string.
        Example: ("Na+", "Cl-") -> "Cl-,Na+"
        """
        sorted_list = sorted([str(x).strip() for x in tuple_pair])
        return ",".join(sorted_list)

    @property
    def param_db(self):
        result = {}
        if self._param_db is None:
            if self.component_groups['cation_anion_pairs']:
                for ca in self.component_groups['cation_anion_pairs']:
                    ca_str = self.order_str(ca)
                    result[ca_str] = db.get_binary(ca_str, self.t)
            if self.component_groups['cation_pairs']:
                for cc in self.component_groups['cation_pairs']:
                    cc_str = self.order_str(cc)
                    result[cc_str] = db.get_binary(cc_str, self.t)
            if self.component_groups['anion_pairs']:
                for aa in self.component_groups['anion_pairs']:
                    aa_str = self.order_str(aa)
                    result[aa_str] = db.get_binary(aa_str, self.t)
            if self.component_groups['nc']:
                for pair in self.component_groups['nc']:
                    pair_str = self.order_str(pair)
                    result[pair_str] = db.get_binary(pair_str, self.t)
            if self.component_groups['na']:
                for pair in self.component_groups['na']:
                    pair_str = self.order_str(pair)
                    result[pair_str] = db.get_binary(pair_str, self.t)
            if self.component_groups['nn']:
                for pair in self.component_groups['nn']:
                    pair_str = self.order_str(pair)
                    result[pair_str] = db.get_binary(pair_str, self.t)
            if self.component_groups['nca']:
                for pair in self.component_groups['nca']:
                    pair_str = self.order_str(pair)
                    result[pair_str ] = db.get_ternary(pair_str, self.t)
            if self.component_groups['cca']:
                for pair in self.component_groups['cca']:
                    pair_str = self.order_str(pair)
                    result[pair_str] = db.get_ternary(pair_str, self.t)
            if self.component_groups['aac']:
                for pair in self.component_groups['aac']:
                    pair_str = self.order_str(pair)
                    result[pair_str] = db.get_ternary(pair_str, self.t)
            self._param_db = result 
        return self._param_db 


    def get_molalities(self, x):
        """
        x = (x1,x2)
        x1: molality of Na
        x2: molality of Cl
        """
        molalities = pm.calculate_molality(tuple(x), self.species)
        return molalities

    def charge_balance(self, x):
        # x = tuple(x)
        # balance = pm.calculate_charge_balance(x, self.get_molalities(x))
        # return balance
        balance = 0
        molalities = self.get_molalities(x)
        for species in molalities.keys():
            balance += get_charge_number(species) * molalities[species]
        return balance

    def get_ionic_strength(self, x):
        """
        For calculating ionic strength. I = (1/2) ∑ mᵢ·zᵢ²
        :param x: a tuple (x1,x2).
        :return: ionic strength.
        """
        molalities = self.get_molalities(x)
        sum_value = 0

        for ion in molalities.keys():
            charge_number = get_charge_number(ion)
            sum_value += molalities[ion] * charge_number ** 2
        return sum_value / 2
    

    def get_z(self, x):
        molalities = self.get_molalities(x)
        ions = molalities.keys()
        z_value = 0
        for ion in ions:
            charge_number = get_charge_number(ion)
            z_value += molalities[ion] * abs(charge_number)
        return z_value


    def get_a_phi(self):
        return pm.a_phi_spencer(self.T)

    def get_b(self, x):
        pair_parameters = {}
        cation_anion_pairs = self.component_groups['cation_anion_pairs']
        for cap in cation_anion_pairs:
            parameters = self.param_db[self.order_str(cap)]
            pair_parameters[cap] = pm.beta_calculate(
                ion_pair=cap,
                ionic_strength=self.get_ionic_strength(x),
                T=self.T,
                parameters = parameters
            )
        return pair_parameters

    def get_b_phi(self, x):
        pair_parameters = {}
        cation_anion_pairs = self.component_groups['cation_anion_pairs']
        for cap in cation_anion_pairs:
            parameters = self.param_db[self.order_str(cap)]
            pair_parameters[cap] = pm.beta_phi_calculate(
                ion_pair=cap,
                ionic_strength=self.get_ionic_strength(x),
                T=self.T,
                parameters = parameters
            )

        return pair_parameters

    def get_b_prime(self, x):

        """
        Get Betas ready for calculating the "F" function.
        :return: betas for each type of salt
        """

        pair_parameters = {}
        cation_anion_pairs = self.component_groups['cation_anion_pairs']

        for cap in cation_anion_pairs:
            parameters = self.param_db[self.order_str(cap)]
            pair_parameters[cap] = pm.beta_prime_calculate(
                ion_pair=cap,
                ionic_strength=self.get_ionic_strength(x),
                T=self.T,
                parameters = parameters
            )
        return pair_parameters

    def get_c(self):
        pair_parameters = {}
        cation_anion_pairs = self.component_groups['cation_anion_pairs']
        for cap in cation_anion_pairs:
            parameters = self.param_db[self.order_str(cap)]
            pair_parameters[cap] = pm.c_calculate(
                ion_pair=cap,
                T=self.T,
                parameters = parameters
            )
        return pair_parameters

    def get_cc_phi(self, x):
        cation_pairs = self.component_groups['cation_pairs']
        phi_dict = {}
        for cap in cation_pairs:
            parameters = self.param_db[self.order_str(cap)]
            phi_dict[cap] = pm.cc_phi_calculate(
                ion_pair=cap,
                ionic_strength=self.get_ionic_strength(x),
                a_phi=self.get_a_phi(),
                T=self.T,
                parameters = parameters
            )
        return phi_dict

    def get_cc_phi_prime_phi(self, x):
        i = self.get_ionic_strength(x)
        dict = {}
        phis = self.get_cc_phi(x)
        for pair in phis.keys():
            dict[pair] = phis[pair]["phi"] + i * phis[pair]["phi_prime"]
        return dict

    def get_aa_phi(self, x):
        if self.component_groups['anion_pairs']:
            anion_pairs = self.component_groups['anion_pairs']
            phi_dict = {}
            for aap in anion_pairs:
                parameters = self.param_db[self.order_str(aap)]
                phi_dict[aap] = pm.aa_phi_calculate(
                    ion_pair=aap,
                    ionic_strength=self.get_ionic_strength(x),
                    a_phi=self.get_a_phi(),
                    T=self.T,
                    parameters=parameters
                )
            return phi_dict
        return {}

    def get_aa_phi_prime_phi(self, x):
        if self.component_groups['anion_pairs']:
            i = self.get_ionic_strength(x)
            dict = {}
            phis = self.get_aa_phi(x)
            for pair in phis.keys():
                dict[pair] = phis[pair]["phi"] + i * phis[pair]["phi_prime"]
            return dict
        else: 
            return {}

    def get_cca_psi(self):
        
        cation_pairs = self.component_groups['cation_pairs']
        anions = self.component_groups['anions']
        cca_pairs = {}
        for cation_pair in cation_pairs:
            cation1 = cation_pair[0]
            cation2 = cation_pair[1]
            for anion in anions:
                parameters = self.param_db[self.order_str((cation1, cation2, anion))]
                ternary_parameters = pm.ternary_parameter_cal(self.T, parameters)
                cca_pairs[cation1, cation2, anion] = ternary_parameters[0]
        return cca_pairs

    def get_aac_psi(self):
        anion_pairs = self.component_groups['anion_pairs']
        if anion_pairs:
            cations = self.component_groups['cations']
            aac_pairs = {}
            for anion_pair in anion_pairs:
                anion_list = list(anion_pair)
                a1 = anion_list[0]
                a2 = anion_list[1]
                for cation in cations:
                    parameters = self.param_db[self.order_str((a1, a2, cation))]
                    ternary_parameters = pm.ternary_parameter_cal(self.T, parameters)
                    aac_pairs[a1, a2, cation] =ternary_parameters[0]
            return aac_pairs
        else:
            return {}

    def get_lambdas(self):
        neutral_ion_pairs = self.component_groups['neutral_ion_pairs']
        lambdas = {}
        for pair in neutral_ion_pairs:
            parameters = self.param_db[self.order_str(pair)]
            lambdas[pair] = pm.lambda_cal(pair, self.T, parameters)
        return lambdas

    def get_zetas(self):
        nca_pairs = self.component_groups['neutral_cation_anion_pairs']
        zetas = {}
        for pair in nca_pairs:
            parameters = self.param_db[self.order_str(pair)]
            zetas[pair] = pm.zeta_cal(pair, self.T, parameters)
        return zetas

    def get_osmotic_coefficient(self, x):
        molalities = self.get_molalities(x)
        m_sum = sum(molalities.values())
        i = self.get_ionic_strength(x)
        a_phi = self.get_a_phi()
        z = self.get_z(x)
        b = self.get_b_phi(x)
        c = self.get_c()
        cations = self.component_groups['cations']
        cation_pairs = self.component_groups['cation_pairs']
        anions = self.component_groups['anions']
        anion_pairs = self.component_groups['anion_pairs']
        cc_phis = self.get_cc_phi_prime_phi(x)
        cca_psis = self.get_cca_psi()

        neutrals = self.component_groups['neutrals']
        lambdas = self.get_lambdas()
        zetas = self.get_zetas()
        nca_pairs = self.component_groups['neutral_cation_anion_pairs']

        item0 = (2 / m_sum)

        item1 = -(a_phi * i ** 1.5) / (1 + 1.2 * i ** 0.5)

        item2 = 0
        for cap in b.keys():
            m_1 = molalities[cap[0]]
            m_2 = molalities[cap[1]]

            b_phi = 0
            for key in b.keys():
                if set(cap) == set(key):
                    b_phi = b[key]

            c_value = 0
            for key in c.keys():
                if set(cap) == set(key):
                    c_value = c[key]

            item2 += m_1 * m_2 * (b_phi + z * c_value)

        item3 = 0
        for cp in cation_pairs:
            c1 = cp[0]
            c2 = cp[1]

            m_c1 = molalities[c1]
            m_c2 = molalities[c2]

            # find value of cc_phi
            cc_phi = 0
            for key in cc_phis.keys():
                if {cp[0], cp[1]} == set(key):
                    cc_phi = cc_phis[key]

            item3_subitem1 = 0
            for anion in anions:
                m_a = molalities[anion]

                # find value of cca_psi
                cca_psi = 0
                for key in cca_psis.keys():
                    if set(key) == {c1, c2, anion}:
                        cca_psi = cca_psis[key]

                item3_subitem1 += m_a * cca_psi

            item3 += m_c1 * m_c2 * (cc_phi + item3_subitem1)

        item4 = 0

        if len(anions) > 1:
            aa_phis = self.get_aa_phi_prime_phi(x)
            aac_psis = self.get_aac_psi()

            for ap in anion_pairs:
                m_a1 = molalities[ap[0]]
                m_a2 = molalities[ap[1]]

                # find value of aa_phi
                aa_phi = 0
                for key in aa_phis.keys():
                    if {ap[0], ap[1]} == set(key):
                        aa_phi = aa_phis[key]

                item4_subitem1 = 0
                for cation in cations:
                    m_c = molalities[cation]

                    aac_psi = 0

                    # find value of aac_psi
                    for key in cca_psis.keys():
                        if {ap[0], ap[1], cation} == set(key):
                            aac_psi = aac_psis[key]

                    item4_subitem1 += m_c * aac_psi
                item4 += m_a1 * m_a2 * (aa_phi + item4_subitem1)

        item5 = 0
        for neutral in neutrals:
            m_n = molalities[neutral]
            for cation in cations:
                m_c = molalities[cation]

                lambda_nc = 0
                for key in lambdas.keys():
                    if {neutral, cation} == set(key):
                        lambda_nc = lambdas[key]
                item5 += m_n * m_c * lambda_nc

        item6 = 0
        for neutral in neutrals:
            m_n = molalities[neutral]
            for anion in anions:
                m_a = molalities[anion]

                lambda_na = 0
                for key in lambdas.keys():
                    if {neutral, anion} == set(key):
                        lambda_na = lambdas[key]
                item6 += m_n * m_a * lambda_na

        item7 = 0
        for nca_pair in nca_pairs:
            m1 = molalities[nca_pair[0]]
            m2 = molalities[nca_pair[1]]
            m3 = molalities[nca_pair[2]]

            zeta_nca = 0
            for key in zetas.keys():
                if set(nca_pair) == set(key):
                    zeta_nca = zetas[key]

            item7 += m1 * m2 * m3 * zeta_nca

        osmotic_coefficient = item0 * (item1 + item2 + item3 + item4 + item5 + item6 + item7) + 1
        return osmotic_coefficient

    def get_water_activity(self, x):
        m_sum = sum(self.get_molalities(x).values())
        phi = self.get_osmotic_coefficient(x)
        ln_activity = -phi * m_sum / 55.50844
        return ln_activity

    def get_f_uppercase(self, x):
        a_phi = self.get_a_phi()
        i = self.get_ionic_strength(x)
        component_groups = self.component_groups
        cation_pairs = component_groups['cation_pairs']
        anion_pairs = component_groups['anion_pairs']
        molalities = self.get_molalities(x)
        ca_beta_primes = self.get_b_prime(x)
        cation_anion_pairs = component_groups['cation_anion_pairs']
        cc_phis = self.get_cc_phi(x)
        aa_phis = self.get_aa_phi(x)
        f_gamma = pm.get_f_gamma(a_phi, i)

        item0 = f_gamma

        item1 = 0
        for cap in cation_anion_pairs:
            m_p1 = molalities[cap[0]]
            m_p2 = molalities[cap[1]]
            ca_beta_prime = 0

            # find the value of beta_prime of cap
            for key in ca_beta_primes.keys():
                if {cap[0], cap[1]} == set(key):
                    ca_beta_prime = ca_beta_primes[key]

            item1 += m_p1 * m_p2 * ca_beta_prime

        item2 = 0
        for cp in cation_pairs:
            m_p1 = molalities[cp[0]]
            m_p2 = molalities[cp[1]]

            cc_phi_prime = 0

            # find the value of beta_prime of cap
            for key in cc_phis.keys():
                if {cp[0], cp[1]} == set(key):
                    cc_phi_prime = cc_phis[key]['phi_prime']

            item2 += m_p1 * m_p2 * cc_phi_prime

        item3 = 0
        if anion_pairs:
            for ap in anion_pairs:
                m_a1 = molalities[ap[0]]
                m_a2 = molalities[ap[1]]

                aa_phi_prime = 0
                for key in aa_phis.keys():
                    if {ap[0], ap[1]} == set(key):
                        aa_phi_prime = aa_phis[key]['phi_prime']
                item3 += m_a1 * m_a2 * aa_phi_prime

        f_uppercase = item0 + item1 + item2 + item3
        return f_uppercase

    def get_cation_activity_coefficients(self, target_cation, x):

        cations = self.component_groups['cations']
        anions = self.component_groups['anions']
        anion_pairs =self.component_groups['anion_pairs']
        cation_anion_pairs = self.component_groups['cation_anion_pairs']
        neutrals = self.component_groups['neutrals']
        molalities = self.get_molalities(x)
        target_cation = target_cation
        betas = self.get_b(x)
        cs = self.get_c()
        z = self.get_z(x)
        lambdas = self.get_lambdas()
        zetas = self.get_zetas()
        cc_phis = self.get_cc_phi(x)
        cation_psis = self.get_cca_psi()
        anion_psis = self.get_aac_psi()

        f_uppercase = self.get_f_uppercase(x)
        charge_number = get_charge_number(target_cation)

        item0 = charge_number ** 2 * f_uppercase

        item1 = 0
        for anion in anions:
            m_a = molalities[anion]
            beta_ca = 0
            for key in betas.keys():
                if {target_cation, anion} == set(key):
                    beta_ca = betas[key]
            c_ca = 0
            for key in cs.keys():
                if {target_cation, anion} == set(key):
                    c_ca = cs[key]
            item1 += m_a * (2 * beta_ca + z * c_ca)

        item2 = 0
        for cation in cations:
            if cation != target_cation:
                m_c = molalities[cation]
                phi_mc = 0
                for key in cc_phis.keys():
                    if set(key) == {cation, target_cation}:
                        phi_mc = cc_phis[key]['phi']

                item2_subitem1 = 0
                for anion in anions:
                    m_a = molalities[anion]
                    for tkey in cation_psis.keys():
                        if {target_cation, cation, anion} == set(tkey):
                            item2_subitem1 += m_a * cation_psis[tkey]
                item2 += m_c * (2 * phi_mc + item2_subitem1)

        item3 = 0
        if anion_pairs:
            for ap in anion_pairs:
                m_a1 = molalities[ap[0]]
                m_a2 = molalities[ap[1]]
                psi_maa = 0

                # find match for psi
                for key in anion_psis.keys():
                    if {ap[0], ap[1], target_cation} == set(key):
                        psi_maa = anion_psis[key]

                item3 += m_a1 * m_a2 * psi_maa

        item4 = 0
        for cap in cation_anion_pairs:
            ion1 = cap[0]
            ion2 = cap[1]
            c_ca = 0

            # find match for c_ca
            for key in cs.keys():
                if set(key) == {ion1, ion2}:
                    c_ca = cs[key]

            item4 += molalities[ion1] * molalities[ion2] * c_ca

        item5 = 0
        for neutral in neutrals:
            m_n = molalities[neutral]

            lambda_nm = 0
            for key in lambdas.keys():
                if {neutral, target_cation} == set(key):
                    lambda_nm = lambdas[key]
            item5 += m_n * lambda_nm

        item6 = 0
        for neutral in neutrals:
            m_n = molalities[neutral]
            for anion in anions:
                m_a = molalities[anion]

                zeta_nam = 0
                for key in zetas.keys():
                    if {target_cation, anion, neutral} == set(key):
                        zeta_nam = zetas[key]
                item6 += m_n * m_a * zeta_nam

        ln_coefficient = item0 + item1 + item2 + item3 + abs(charge_number) * item4 + 2 * item5 + item6
        return ln_coefficient

    def get_anion_activity_coefficients(self, target_anion, x):
        groups = self.component_groups
        cations = groups['cations']
        anions = groups['anions']
        cation_pairs = groups['cation_pairs']
        cation_anion_pairs = groups['cation_anion_pairs']
        neutrals = groups['neutrals']
        lambdas = self.get_lambdas()
        zetas = self.get_zetas()
        molalities = self.get_molalities(x)
        target_anion = target_anion
        betas = self.get_b(x)
        cs = self.get_c()
        z = self.get_z(x)

        aa_phis = self.get_aa_phi(x)

        aac_psis = self.get_aac_psi()
        cca_psis = self.get_cca_psi()

        f_uppercase = self.get_f_uppercase(x)
        charge_number = get_charge_number(target_anion)

        item0 = charge_number ** 2 * f_uppercase

        item1 = 0
        for cation in cations:
            m_a = molalities[cation]
            beta_ca = 0
            for key in betas.keys():
                if {target_anion, cation} == set(key):
                    beta_ca = betas[key]
            c_ca = 0
            for key in cs.keys():
                if {target_anion, cation} == set(key):
                    c_ca = cs[key]
            item1 += m_a * (2 * beta_ca + z * c_ca)

        item2 = 0
        for anion in anions:
            if anion != target_anion:
                m_a = molalities[anion]
                phi_xa = 0
                for key in aa_phis.keys():
                    if set(key) == {anion, target_anion}:
                        phi_xa = aa_phis[key]['phi']

                item2_subitem1 = 0
                for cation in cations:
                    m_c = molalities[cation]
                    for tkey in aac_psis.keys():
                        if {target_anion, cation, anion} == set(tkey):
                            item2_subitem1 += m_c * aac_psis[tkey]
                item2 += m_a * (2 * phi_xa + item2_subitem1)

        item3 = 0
        for cp in cation_pairs:
            m_c1 = molalities[cp[0]]
            m_c2 = molalities[cp[1]]
            psi_ccx = 0

            # find match for psi
            for key in cca_psis.keys():
                if {cp[0], cp[1], target_anion} == set(key):
                    psi_ccx = cca_psis[key]
            item3 += m_c1 * m_c2 * psi_ccx

        item4 = 0
        for cap in cation_anion_pairs:
            ion1 = cap[0]
            ion2 = cap[1]
            m1 = molalities[ion1]
            m2 = molalities[ion2]

            c_ca = 0
            for key in cs.keys():
                if set(key) == {ion1, ion2}:
                    c_ca = cs[key]

            item4 += m1 * m2 * c_ca

        item5 = 0
        for neutral in neutrals:
            m_n = molalities[neutral]
            lambda_nx = 0

            for key in lambdas.keys():
                if {neutral, target_anion} == set(key):
                    lambda_nx = lambdas[key]

            item5 += m_n * lambda_nx

        item6 = 0
        for neutral in neutrals:
            m_n = molalities[neutral]
            for cation in cations:
                m_c = molalities[cation]

                zeta_nam = 0

                for key in zetas.keys():
                    if {target_anion, cation, neutral} == set(key):
                        zeta_nam = zetas[key]
                item6 += m_n * m_c * zeta_nam

        ln_coefficient = item0 + item1 + item2 + item3 + abs(charge_number) * item4 + 2 * item5 + item6

        return ln_coefficient

    def get_neutral_activity_coefficients(self, target_neutral, x):
        groups = self.component_groups
        cations = groups['cations']
        anions = groups['anions']
        cation_anion_pairs = groups['cation_anion_pairs']

        molalities = self.get_molalities(x)

        lambdas = self.get_lambdas()
        zetas = self.get_zetas()

        item0 = 0
        for cation in cations:
            m_c = molalities[cation]
            lambda_nc = 0
            for key in lambdas.keys():
                if {cation, target_neutral} == set(key):
                    lambda_nc = lambdas[key]
            item0 += m_c * lambda_nc

        item1 = 0
        for anion in anions:
            m_a = molalities[anion]
            lambda_na = 0
            for key in lambdas:
                if {anion, target_neutral} == set(key):
                    lambda_na = lambdas[key]
            item1 += m_a * lambda_na

        item2 = 0
        for pair in cation_anion_pairs:
            m1 = molalities[pair[0]]
            m2 = molalities[pair[1]]

            zeta_nca = 0
            for key in zetas.keys():
                if {target_neutral, pair[0], pair[1]} == set(key):
                    zeta_nca = zetas[key]

            item2 += m1 * m2 * zeta_nca

        ln_coefficient = 2 * item0 + 2 * item1 + item2

        return ln_coefficient

    def total_gibbs_energy(self, x):
        r = 8.314
        molalities = self.get_molalities(x)
        total_g = 0
        for species in molalities.keys():
            species_type = type_of_species(species)
            m_i = molalities[species]
            if m_i != 0:
                ln_gamma = 0
                if species_type == 0:
                    ln_gamma = self.get_neutral_activity_coefficients(species, x)
                elif species_type == 1:
                    ln_gamma = self.get_cation_activity_coefficients(species, x)
                elif species_type == -1:
                    ln_gamma = self.get_anion_activity_coefficients(species, x)
                total_g += m_i * (0 + r * self.T * (np.log(m_i) + ln_gamma))

        # gibbs energy of water
        ln_a_w = self.get_water_activity(x)

        water_data = db.get_reaction("H2(g) + 0.5O2(g) = H2O(l)", self.t)
        water_parameters = water_data['analytic']

        std_cp_water = pm.get_chemical_potential(water_parameters, self.T, water_data['eq_num'])

        cp_water = (1000 / 18.015) * (std_cp_water + r * self.T * ln_a_w)

        total_g = total_g + cp_water
        return total_g

    def solubility_equilibrium(self, x):
        # print(x)
        lna_pitzer = self.get_water_activity(x)
        if self.equilibrium == 'H2O(s) = H2O(l)':
            lnk_ice = clegg_and_brimblecombe(self.T)
            # lnk_ice = monnin(self.t)
            # lnk_ice = spencer(self.t)
            f = lnk_ice - lna_pitzer
        else:
            solid_data = self.solid_disso['stoich']
            parameters = self.solid_disso["analytic"]

            molalities = self.get_molalities(x)
            lnk_potential = pm.get_chemical_potential(parameters, self.T, self.solid_disso["eq_num"])

            lnk_activity = 0
            for species in solid_data.keys():
                # get the stochiometric number of this species first
                sto = solid_data[species]['value']

                if solid_data[species]['type'] == 'cation':
                    m_c = molalities[species]
                    ln_gamma_c = self.get_cation_activity_coefficients(species, x)
                    lnk_activity += sto * (np.log(m_c) + ln_gamma_c)
                elif solid_data[species]['type'] == 'anion':
                    m_a = molalities[species]
                    ln_gamma_a = self.get_anion_activity_coefficients(species, x)
                    lnk_activity += sto * (np.log(m_a) + ln_gamma_a)
                else:
                    lnk_activity += sto * lna_pitzer

            # if a neutral species should be considered.
            if self.neutral:
                ln_gamma_n = self.get_neutral_activity_coefficients(target_species, x)

                lnk_activity = (np.log(molalities[target_species]) + ln_gamma_n) - lnk_activity
            f = lnk_potential - lnk_activity
        return f

    def solve(self):
        # Start the timer
        start_time = time.time()

        bounds = ((0, None), (0, None))
        # bounds = ((0, None))
        cons1 = {'type': 'eq', 'fun': self.charge_balance}
        cons2 = {'type': 'eq', 'fun': self.solubility_equilibrium}
        res = minimize(
            fun=self.total_gibbs_energy,
            x0=self.x0,
            method='SLSQP',
            bounds=bounds,
            constraints=(cons1, cons2)
            # constraints=(cons1)
        )

        # end of timer
        elapsed_time = time.time() - start_time
        # print('times used:', elapsed_time)

        if res.success:
            return res.x
        else:
            return [None, None]



from pypitzer.Pitzer.models import FluidPitzer

species = {
    'Na+': 1, # always be 1 if Na is the internal standard
    'K+': 2 ,  # K/Na = 2
    # 'Mg+2':0.1
}

# create a fluid object with information from microthemometric and LA-ICP-MS data
fluid = FluidPitzer(
    # the initial guess
    x0=(1,1),
  
    # species defined before
    species=species,
  
    # the last melting solid phase
    equilibrium='KCl(s) = K+(aq) + Cl-(aq)',
  
    # melting temperature of the last-melting solid phase, [°C]
    t = 25,
)

result = fluid.solve()

print(result)
# [1.84525981 5.53577942] or [None, None] when there's no solution



import sqlite3
from pathlib import Path
import hashlib

current_dir = Path(__file__).parent
json_path = current_dir / "pypitzer_reaction.json"

db_path = current_dir / "pypitzer.sqlite"

# --- canonicalize reaction string ---
def preprocess_reaction(reaction: str) -> str:
    lhs, rhs = reaction.split("=")
    lhs_list = [s.strip() for s in lhs.split(" + ")]
    rhs_list = [s.strip() for s in rhs.split(" + ")]
    lhs_sorted = " + ".join(sorted(lhs_list))
    rhs_sorted = " + ".join(sorted(rhs_list))
    return f"{lhs_sorted} = {rhs_sorted}"

def reaction_hash(reaction: str) -> str:
    return hashlib.sha256(reaction.encode()).hexdigest()
    
def insert_reaction(conn: sqlite3.Connection,
                    reaction: str,
                    analytic: list[float],
                    eq_num: int,
                    ref: str,
                    T_min: float,
                    T_max: float):
    """
    Insert a single reaction into the 'reaction' table.

    Parameters
    ----------
    conn : sqlite3.Connection
        Open connection to the SQLite database.
    reaction : str
        Reaction string, e.g., "H2O = H+ + OH-".
    analytic : list of float
        Coefficients for the analytic lnK expression.
    eq_num : int
        Equation number / model identifier.
    ref : str
        Reference for the data.
    T_min : float
        Minimum temperature (inclusive).
    T_max : float
        Maximum temperature (exclusive).
    """
    

    reaction_canon = preprocess_reaction(reaction)
    r_hash = reaction_hash(reaction_canon)
    
    # Convert analytic list to comma-separated string
    analytic_str = ",".join(str(x) for x in analytic)
    
    cur = conn.cursor()
    try:
        cur.execute("""
            INSERT INTO reaction (reaction_id, reaction, analytic, eq_num, ref, T_min, T_max)
            VALUES (?, ?, ?, ?, ?, ?, ?)
        """, (r_hash, reaction_canon, analytic_str, eq_num, ref, T_min, T_max))
        conn.commit()
    except sqlite3.IntegrityError as e:
        raise ValueError(f"Failed to insert reaction: {e}")



if __name__ == "__main__":
    conn = sqlite3.connect(db_path)

    data =[
            {
            "reaction": "LiCl(aq) = Li+(aq) + Cl-(aq)",
            "analytic": [-3.5869,0.007566,-1254.4,0,0,],
            "T_min": 25,
            "T_max": 300,
            },
            {
            "reaction": "LiCl-H2O(s) = Li+(aq) + Cl-(aq) + H2O(l)",
            "analytic": [36.2702,0.005655998,2.696341,-13.2627354,0,],
            "T_min": 25,
            "T_max": 300,
            },
            {
            "reaction": "LiCl-2H2O(s) = Li+(aq) + Cl-(aq) + 2H2O(l)",
            "analytic":[6.1532,-0.0064179,0,0,0,],
            "T_min": 25,
            "T_max": 300,
            },
            {
            "reaction": "KCl(s) = K+(aq) + Cl-(aq)",
            "analytic": [6.496259873, -0.012323368, -1537.997915, 1.507740146, -42632.32102],
            "T_min": 25,
            "T_max": 300,
            },
            {
            "reaction": "NaCl(s) = Na+(aq) + Cl-(aq)",
            "analytic": [-752.24954, -0.11904958, 41385.703, 274.17933, -2480.9109e+03],
            "T_min": 25,
            "T_max": 300,
            },
            {
            "reaction": "LiOH-H2O(s) = Li+(aq) + OH-(aq) + H2O(l)",
            "analytic": [-3.8702256, -0.00914328, 0.4129607, 3.0072101, 0],
            "T_min": 25,
            "T_max": 200,
            },
            {
            "reaction": "LiOH(s) = Li+(aq) + OH-(aq)",
            "analytic": [-256.08895, -0.06047962, 9290.232715, 98.8913449, 142.35284],
            "T_min": 50,
            "T_max": 200,
            },
    ]

    for d in data:
        insert_reaction(
            conn=conn,
            reaction= d['reaction'],
            analytic= d['analytic'],
            eq_num= 5,
            ref= "Lassin et al., 2015",
            T_min= d['T_min'],
            T_max= d['T_max'],
        )



# -*- coding: utf-8 -*-
# Author: Yiping Liu
# Description: This script calculates the average of a list of numbers.
# Version: 1.0
# Last Modified: Nov 11, 2025
import json
import re
import numpy as np
import pandas as pd
import itertools

from pypitzer.public.j_x import compute_j_jp
from pypitzer.public.low_level import get_charge_number


def parameter_fun_spencer(a, T):
    # eq_num = 0
    return a[0] + a[1] * T + a[2] * T ** 2 + a[3] * T ** 3 + a[4] / T + a[5] * np.log(T)

def parameter_fun_marion(a, T):
    # eq_num = 1
    return  a[0] + a[1] * T + a[2] * T ** 2 + a[3] * T ** 3 + a[4] / T + a[5] * np.log(T) + a[6] / (T ** 2)

def parameter_fun_eq2(a, T):
    # eq_num = 2
    return a[1] + a[2] * T + a[3] * T ** 2 + a[4] * T ** 3 + a[5] / T + a[6] * np.log(T) + a[7 ]/ (T - 263) + a[8] / (680 - T)
    
def parameter_fun_moller(a, T):
    # eq_num = 3
    return a[0] + a[1] * T + a[2] / T + a[3] * np.log(T) + a[4] / (T - 263) + a[5]* T ** 2 + a[6] / (680 - T) + a[7] / (T - 227)

def parameter_fun_holmes(a, T):
    # eq_num = 4
    Tr = 298.15
    return a[1] + a[2]*(1/T - 1/Tr) + a[3]*np.log(T/Tr) + a[4]*(T - Tr) + a[5]*(T**2 - Tr**2) + a[6]*np.log(T - 260)
     
def parameter_fun_eq5(a, T):
    # eq_num = 5
    log_k = a[0] + a[1] * T + a[2] / T + a[3] * np.log10(T) + a[4] / T**2
    return log_k * 2.303


def get_parameter_pypitzer(a, T, eq_num):
    if eq_num == 0:
        return parameter_fun_spencer(a, T)
    elif eq_num == 1:
        return parameter_fun_marion(a, T)
    elif eq_num == 2:
        return parameter_fun_eq2(a, T)
    elif eq_num == 3:
        return parameter_fun_moller(a, T)
    elif eq_num == 4:
        return parameter_fun_holmes(a, T)
    elif eq_num == 5:
        return parameter_fun_eq5(a, T)



def a_phi_moller(T):
    a = [
         3.36901532e-01,
        -6.32100430e-04,
         9.142523590e00,
        -1.35143986e-02,
         2.26089488e-03,
         1.92118597e-06,
         4.525864640e01,
                      0,
    ]
    return parameter_fun_moller(a, T)


def a_phi_spencer(T):
    """
    Calculate the A(φ) (Debye-Hukel constant) with temperature-dependent equation.
    :param T: temperature of solution, [K]
    :return: A(φ)
    """
    a = [
            8.66836498e1,
            8.48795942e-2,
           -8.88785150e-5,
            4.88096393e-8,
           -1.32731477e3,
           -1.76460172e1
        ]
    return parameter_fun_spencer(a, T)


def g_func(a):
    result = 2 * (
            1 - (1 + a) * np.exp(-a)
    ) / a ** 2
    return result


def g_func_prime(a):
    g_prime = -2 * (
            1 - (1 + a + a ** 2 / 2) * np.exp(-a)
    ) / a ** 2
    return g_prime


def get_beta(parameters, pair, i):
    z1 = get_charge_number(pair[0])
    z2 = get_charge_number(pair[1])

    alpha = 2  # kg^(1/2)⋅mol^(-1/2)
    alpha_1 = 1.4  # kg^(1/2)⋅mol^(-1/2)
    alpha_2 = 12  # kg^(1/2)⋅mol^(-1/2)
    b0, b1, b2 = parameters

    if abs(z1) == 2 and abs(z2) == 2:
        beta_mx = b0 + b1 * g_func(
            alpha_1 * (i ** (1 / 2))
        ) + b2 * g_func(
            alpha_2 * (i ** (1 / 2))
        )
    else:
        beta_mx = b0 + b1 * g_func(
            alpha * i ** (1 / 2)
        )
    return beta_mx


def get_beta_prime(parameters, pair, i):
    charge_number1 = get_charge_number(pair[0])
    charge_number2 = get_charge_number(pair[1])

    alpha = 2  # kg^(1/2)⋅mol^(-1/2)
    alpha_1 = 1.4  # kg^(1/2)⋅mol^(-1/2)
    alpha_2 = 12  # kg^(1/2)⋅mol^(-1/2)
    b0, b1, b2 = parameters
    if abs(charge_number1) == 2 and abs(charge_number2) == 2:
        beta_prime = (
                             b1 * g_func_prime(
                         alpha_1 * (i ** (1 / 2))
                     ) + b2 * g_func_prime(
                         alpha_2 * (i ** (1 / 2))
                     )
                     ) / i
    else:
        beta_prime = b1 * g_func_prime(alpha * i ** (1 / 2)) / i
    return beta_prime


def get_beta_phi(parameters, pair, i):
    z1 = get_charge_number(pair[0])
    z2 = get_charge_number(pair[1])

    alpha = 2  # kg^(1/2)⋅mol^(-1/2)
    alpha_1 = 1.4  # kg^(1/2)⋅mol^(-1/2)
    alpha_2 = 12  # kg^(1/2)⋅mol^(-1/2)

    b0, b1, b2 = parameters

    """for 2-2 type salts"""
    if abs(z1) == 2 and abs(z2) == 2:
        beta_phi = b0 + b1 * np.exp(-alpha_1 * (i ** (1 / 2))) + b2 * np.exp(-alpha_2 * (i ** (1 / 2)))
    else:
        """for 1-1, 1-2, 2-1, 3-1, 4-1 type salts"""
        beta_phi = b0 + b1 * np.exp(-alpha * (i ** (1 / 2)))
    return beta_phi

def get_c(c_phi, z_m, z_x):
    c_mx = c_phi / (2 * (abs(z_m * z_x)) ** 0.5)
    return c_mx

def get_c_gamma(c_phi):
    c_gamma = 3 * c_phi / 2
    return c_gamma


def get_f(a_phi, i):
    """
    :param a_phi: A_phi (Debye-Hukel constant)
    :param i: ionic strength
    :return: expression of "f" function
    """
    b = 1.2
    f = - (4 * i * a_phi / b) * np.log(1 + b * i ** (1 / 2))
    return f


def get_f_gamma(a_phi, i):
    """
    :param a_phi:A_phi (Debye-Hukel constant)
    :param i: ionic strength
    :return: the "f^gamma" function in Pitzer's model
    """
    b = 1.2
    f_gamma = - a_phi * (
            i ** (1 / 2) / (1 + b * i ** (1 / 2)) + (2 / b) * np.log(
        1 + b * i ** (1 / 2))
    )
    return f_gamma


"""
*** End of the Pitzer parameters dealing *** 
"""


def get_x_mn(z_m: int, z_n: int, a_phi: float, i: float):
    """
    calculate the 'x' value of ions 'm' and 'n'.
    :param z_m: charge number of ion 'm'
    :param z_n: charge number of ion 'n'
    :param a_phi: Avogedral's number of this solution
    :param i: Ionic strength of this solution
    :return: 'x_mn' for calculating the 'J' value
    :reference: [2] p9
    """
    x_mn = 6 * z_m * z_n * a_phi * i ** 0.5
    return x_mn


def get_e_theta(z_m, z_n, a_phi, i):
    """
    :param z_m: charge number of species m
    :param z_n: charge number of species n
    :param a_phi:
    :param i: ionic strength
    :return: e_theta and e_theta_prime
    :reference: [1] p123
    """

    x_mn = get_x_mn(z_m, z_n, a_phi, i)
    x_mm = get_x_mn(z_m, z_m, a_phi, i)
    x_nn = get_x_mn(z_n, z_n, a_phi, i)

    mn = compute_j_jp(x_mn)
    mm = compute_j_jp(x_mm)
    nn = compute_j_jp(x_nn)
    j_mn = mn['j_x']
    j_mn_prime = mn['j_xp']
    j_mm = mm['j_x']
    j_mm_prime = mm['j_xp']
    j_nn = nn['j_x']
    j_nn_prime = nn['j_xp']

    e_theta = (z_m * z_n / (4 * i)) * (j_mn - 0.5 * j_mm - 0.5 * j_nn)
    e_theta_prime = -(e_theta / i) + (z_m * z_n / (8 * i ** 2)) * (
            x_mn * j_mn_prime - 0.5 * x_mm * j_mm_prime - 0.5 * x_nn * j_nn_prime)
    return {
        "e_theta": e_theta,
        "e_theta_prime": e_theta_prime,
    }


def calculate_ionic_strength(molalities):
    data = molalities
    ions = data.keys()
    sum_value = 0
    for ion in ions:
        charge_number = get_charge_number(ion)
        sum_value += data[ion] * (charge_number ** 2)
    return sum_value / 2


def calculate_molality(x, species):
    x1, x2 = x
    molalities = {}
    for key, value in species.items():
        if key != 'Cl-':
            molalities[key] = value * x1
    molalities['Cl-'] = x2
    return molalities



def calculate_charge_balance(x, molalities):
    balance = 0
    for species in molalities.keys():
        balance += get_charge_number(species) * molalities[species]
    return balance


def get_chemical_potential(parameters, T, eq_num):
    if eq_num == 0:
        return parameter_fun_spencer(parameters, T)
    else:
        return 0


import itertools

def order_tuple(t):
    """
    Order elements of a tuple alphabetically.
    Example: ("Na+", "Cl-") -> ("Cl-", "Na+")
    """
    return tuple(sorted([str(x).strip() for x in t]))

def group_components(components):
    """
    Find groups from components of ions and neutral species.
    :param components: list of cations, anions, and neutral species.
    :return: dictionary of grouped components with tuples
    """
    cations = [c for c in components if '+' in c]
    anions = [a for a in components if '-' in a]
    neutrals = [n for n in components if '+' not in n and '-' not in n]

    # Basic pairs
    cation_anion_pairs = [order_tuple(p) for p in itertools.product(cations, anions)]
    cation_pairs = [order_tuple(p) for p in itertools.combinations(cations, 2)] if len(cations) >= 2 else []
    anion_pairs = [order_tuple(p) for p in itertools.combinations(anions, 2)] if len(anions) >= 2 else []
    neutral_pairs = [order_tuple(p) for p in itertools.combinations(neutrals, 2)]
    neutral_ion_pairs = [order_tuple(p) for p in itertools.product(neutrals, cations + anions)]
    neutral_cation_anion_pairs = [order_tuple((n, c, a)) for n in neutrals for c, a in itertools.product(cations, anions)]

    # Additional triplets
    cca = [order_tuple((c1, c2, a)) for c1, c2 in itertools.combinations(cations, 2) for a in anions]
    aac = [order_tuple((a1, a2, c)) for a1, a2 in itertools.combinations(anions, 2) for c in cations]

    # Neutral-cation / neutral-anion pairs
    nc = [order_tuple((n, c)) for n in neutrals for c in cations]
    na = [order_tuple((n, a)) for n in neutrals for a in anions]
    nca = [order_tuple((n, c, a)) for n in neutrals for c, a in itertools.product(cations, anions)]

    return {
        'cations': tuple(cations),
        'anions': tuple(anions),
        'neutrals': tuple(neutrals),
        'cation_anion_pairs': tuple(cation_anion_pairs),
        'cation_pairs': tuple(cation_pairs),
        'anion_pairs': tuple(anion_pairs),
        'nn': tuple(neutral_pairs),
        'neutral_ion_pairs': tuple(neutral_ion_pairs),
        'neutral_cation_anion_pairs': tuple(neutral_cation_anion_pairs),
        'cca': tuple(cca),
        'aac': tuple(aac),
        'nc': tuple(nc),
        'na': tuple(na),
        'nca': tuple(nca)
    }




def ternary_parameter_cal(T, parameters):
    para_psi  = parameters['psi']
    para_zeta = parameters['zeta'] if 'zeta' in parameters else None
    eq_num = parameters['eq_num']

    psi = get_parameter_pypitzer(para_psi, T, eq_num)
    zeta = get_parameter_pypitzer(para_zeta, T, eq_num) if para_zeta else 0
    return (psi, zeta)


def get_beta_012(ion_pair, T, parameters):
    para_b0 = parameters['b0']
    para_b1 = parameters['b1']
    para_b2 = parameters['b2'] if "b2" in parameters else None
    eq_num = parameters['eq_num']

    b0 = get_parameter_pypitzer(para_b0, T, eq_num) 
    b1 = get_parameter_pypitzer(para_b1, T, eq_num)
    b2 = get_parameter_pypitzer(para_b2, T, eq_num) if para_b2 else 0
    return (b0, b1, b2)


def beta_calculate(ion_pair, ionic_strength, T, parameters):
    beta_012 = get_beta_012(ion_pair, T, parameters)
    b_phi = get_beta(beta_012, ion_pair, ionic_strength)
    return b_phi


def beta_phi_calculate(ion_pair, ionic_strength, T, parameters):
    beta_012 = get_beta_012(ion_pair, T, parameters)
    b_phi = get_beta_phi(beta_012, ion_pair, ionic_strength)
    return b_phi

def beta_prime_calculate(ion_pair, ionic_strength, T, parameters):
    beta_012 = get_beta_012(ion_pair, T, parameters)
    b_prime = get_beta_prime(beta_012, ion_pair, ionic_strength)
    return b_prime

def c_calculate(ion_pair, T, parameters):
    para_cphi = parameters['c_phi']

    eq_num = parameters['eq_num']
    c0 = get_parameter_pypitzer(para_cphi, T, eq_num)

    z1 = get_charge_number(ion_pair[0])
    z2 = get_charge_number(ion_pair[1])

    c = get_c(c0, z1, z2)
    return c

def lambda_cal(pair,T, parameters):
    # parameters = binary_query(pair)

    para_lambda = parameters['lambda']
    eq_num = parameters['eq_num']
    return get_parameter_pypitzer(para_lambda, T, eq_num)


def zeta_cal(pair,T, parameters):
    # parameters = binary_query(pair)
    para_zeta = parameters['zeta']
    eq_num = parameters['eq_num']
    return get_parameter_pypitzer(para_zeta, T, eq_num)

def cc_phi_calculate(ion_pair, a_phi, ionic_strength, T, parameters):
    z1 = get_charge_number(ion_pair[0])
    z2 = get_charge_number(ion_pair[1])

    para_theta = parameters['theta']
    eq_num = parameters['eq_num']

    theta = get_parameter_pypitzer(para_theta, T, eq_num)

    if z1 != z2:
        e_thetas = get_e_theta(z1, z2, a_phi, ionic_strength)
        e_theta = e_thetas['e_theta']
        e_theta_prime = e_thetas['e_theta_prime']
    else:
        e_theta = 0
        e_theta_prime = 0

    phi = theta + e_theta
    phi_prime = e_theta_prime

    return {
        'phi': phi,
        'phi_prime': phi_prime
    }


def aa_phi_calculate(ion_pair, a_phi, ionic_strength, T, parameters):
    z1 = get_charge_number(ion_pair[0])
    z2 = get_charge_number(ion_pair[1])

    # parameters = binary_query(ion_pair)

    para_theta = list(parameters['theta'].values())
    eq_num = parameters['eq_num']

    theta = get_parameter_pypitzer(para_theta, T, eq_num)
    if z1 != z2:
        e_thetas = get_e_theta(z1, z2, a_phi, ionic_strength)
        e_theta = e_thetas['e_theta']
        e_theta_prime = e_thetas['e_theta_prime']
    else:
        e_theta = 0
        e_theta_prime = 0

    phi = theta + e_theta
    phi_prime = e_theta_prime

    return {
        "phi": phi,
        "phi_prime": phi_prime
    }


"""
***************************
 DATA OF SOLID PHASE
***************************
"""

def clean_state_marks(text: str) -> str:
    """Remove state marks like (aq), (s), (l), (g)."""
    return re.sub(r'\([a-zA-Z]+\)', '', text)


def parse_species(species_str: str):
    """
    Parse species like '2K+' or 'Fe+2' or 'H2O' to get value and type.
    Removes state marks beforehand.
    """
    s = species_str.strip()
    if not s:
        raise ValueError("Empty species string")

    m = re.match(r'^(\d+)\s*(.+)$', s)
    if m:
        coef = int(m.group(1))
        name = m.group(2).strip()
    else:
        coef = 1
        name = s

    if '+' in name and '-' not in name:
        type_ = 'cation'
    elif '-' in name and '+' not in name:
        type_ = 'anion'
    elif '+' in name and '-' in name:
        if name.endswith('+'):
            type_ = 'cation'
        elif name.endswith('-'):
            type_ = 'anion'
        else:
            type_ = 'neutral'
    else:
        type_ = 'neutral'

    return name, {'value': coef, 'type': type_}


def get_solid_stoichiometry(reaction_data):
    """
    Return only the RHS species dictionary (no outer solid key).
    """
    reaction_str = reaction_data['reaction']
    stoich = {}
    if reaction_str:
        lhs, rhs = reaction_str.split('=', 1)

        rhs_no_states = clean_state_marks(rhs)
        species_tokens = [tok.strip() for tok in re.split(r'\s+\+\s+', rhs_no_states.strip()) if tok.strip()]

        
        for tok in species_tokens:
            name, info = parse_species(tok)
            stoich[name] = info
    return stoich


# class Reaction:
#     def __init__(self, reaction_key):
#         self.reaction_key = reaction_key

#     def reaction_data(self):
#         if self.reaction_key not in reaction_database:
#             raise ValueError(f"{self.reaction_key} not found in reactions dictionary")
#         return reaction_database[self.reaction_key]

#     @property
#     def stoich(self):
#         return get_solid_stoichiometry(self.reaction_data())
    
#     @property
#     def analytic(self):
#         return self.reaction_data()['analytic']
    
#     @property
#     def eq_num(self):
#         return self.reaction_data()['eq_num']



"""
Reference
[1] Pitzer K S. Activity coefficients in electrolyte solutions[M]. CRC press, 2018.
[2] Pitzer K S. Thermodynamics of electrolytes. V. Effects of higher-order electrostatic terms[J]. Journal of Solution 
Chemistry, 1975, 4(3): 249-265.
[3] Bradley, Daniel J., and Kenneth S. Pitzer. "Thermodynamics of electrolytes. 12. Dielectric properties of water and 
Debye-Hueckel parameters to 350. degree. C and 1 kbar." Journal of physical chemistry 83.12 (1979): 1599-1603.
[4] Marion GM, Catling DC, Kargel JS. Modeling aqueous ferrous iron chemistry at low temperatures with application to 
Mars. Geochimica et cosmochimica Acta. 2003 Nov 15;67(22):4251-66.
"""



import sqlite3
import json
from pathlib import Path
current_dir = Path(__file__).parent

def preprocess_pair(pair: str) -> str:
    """
    Preprocess a pair string:
    1. Remove spaces
    2. Order ions alphabetically
    """
    ions = [ion.strip() for ion in pair.split(",")]  # remove spaces
    ions.sort()  # alphabetical order
    return ",".join(ions)

def preprocess_data(parmeters: str) -> str:
    """
    Preprocess a pair string:
    1. Remove spaces
    2. Order ions alphabetically
    """
    cleaned_parameters = parmeters.strip() # remove spaces
    return cleaned_parameters 

def preprocess_triplet(triplet: str) -> str:
    """
    Preprocess a ternary triplet string:
    1. Remove spaces
    2. Order ions alphabetically
    """
    ions = [ion.strip() for ion in triplet.split(",")]
    ions.sort()
    return ",".join(ions)


def insert_ternary_param(conn, triplet, ref, T_range, eq_num, psi=None, zeta=None):
    """
    Insert a single ternary parameter into the ternary_params table.

    Parameters:
        conn: sqlite3.Connection
        triplet: str, e.g., "Cl-,K+,Na+"
        ref: str, reference
        T_range: list or tuple of [T_min, T_max]
        eq_num: int, equation index
        psi, zeta: str or None, comma-separated 6 coefficients
            If None, defaults to "0,0,0,0,0,0"
    """
    default_coeff = "0,0,0,0,0,0"
    triplet = preprocess_triplet(triplet)

    cur = conn.cursor()
    cur.execute("""
        INSERT OR IGNORE INTO ternary_params
        (triplet, ref, T_min, T_max, eq_num, psi, zeta)
        VALUES (?, ?, ?, ?, ?, ?, ?)
    """, (
        triplet,
        ref,
        T_range[0],
        T_range[1],
        eq_num,
        preprocess_data(psi) if psi is not None else default_coeff,
        preprocess_data(zeta) if zeta is not None else default_coeff,
    ))
    conn.commit()
    print(f"Successfully inserted/ignored ternary parameter for '{triplet}'")


if __name__ == "__main__":
    db_path = current_dir / "pypitzer.sqlite"
    conn = sqlite3.connect(db_path)




    data = {
        "CO2,Fe+2,Cl-":     "-1.34260256e3,-7.72286e-1,3.91603e-4,0,2.772680974e4,2.5362319406e2,0",
        "CO2,Fe+2,SO4-2":   "-7.37424392e3,-4.608331,2.489207e-3,0,1.431626076e5,1.412302898e5,0",
        "O2,Na+,Cl-":       "0,0,0,0,-2.739,0,0",
        "O2,Na+,OH-":       "-1.25e-2,0,0,0,0,0,0",
        "O2,Na+,NO3-":      "-1.20e-2,0,0,0,0,0,0",
        "O2,Na+,HCO3-":     "0.0,0,0,0,0,0,0",
        "O2,Na+,CO3-2":     "-1.81e-2,0,0,0,0,0,0",
        "O2,Na+,SO4-2":     "-4.60e-2,0,0,0,0,0,0",
        "O2,K+,Cl-":        "-2.11e-2,0,0,0,0,0,0",
        "O2,K+,OH-":        "2.342e-3,0,0,0,0,0,-8.3615e2",
        "O2,K+,NO3-":       "-2.81e-2,0,0,0,0,0,0",
        "O2,K+,SO4-2":      "0.0,0,0,0,0,0,0",
        "O2,Mg+2,Cl-":      "-5.65e-3,0,0,0,0,0,0",
        "O2,Mg+2,SO4-2":    "0,0,0,0,0,0,0",
        "O2,Ca+2,Cl-":      "-1.69e-2,0,0,0,0,0,0",
        "O2,Ca+2,NO3-":     "0,0,0,0,0,0,0",
        "O2,Fe+2,Cl-":      "-5.65e-3,0,0,0,0,0,0",
        "O2,Fe+2,SO4-2":    "0,0,0,0,0,0,0",
        "O2,H+,Cl-":        "-7.7e-3,0,0,0,0,0,0",
    }

    for k, v in data.items():
        insert_ternary_param(
            conn,
            triplet=k, 
            ref="Marion., 2003",
            T_range=[-90, 25],
            eq_num=1,
            psi=None,
            zeta=v
        )

    conn.close()


import sqlite3
import json
from pathlib import Path
current_dir = Path(__file__).parent

def preprocess_pair(pair: str) -> str:
    """
    Preprocess a pair string:
    1. Remove spaces
    2. Order ions alphabetically
    """
    ions = [ion.strip() for ion in pair.split(",")]  # remove spaces
    ions.sort()  # alphabetical order
    return ",".join(ions)

def preprocess_data(parmeters: str) -> str:
    """
    Preprocess a pair string:
    1. Remove spaces
    2. Order ions alphabetically
    """
    cleaned_parameters = parmeters.strip() # remove spaces
    return cleaned_parameters 

def insert_binary_param(conn, pair, ref, T_range, eq_num,
                        b0=None, b1=None, b2=None, c_phi=None, lambda_=None, theta=None):
    """
    Insert a single binary parameter into the binary_params table.
    
    Parameters:
        conn: sqlite3.Connection
        pair: str, e.g., "Cl-,Na+"
        ref: str, reference
        T_range: list or tuple of [T_min, T_max]
        eq_num: int, equation index
        b0, b1, b2, c_phi, lambda_, theta: str or None, comma-separated 6 coefficients
            If None, defaults to "0,0,0,0,0,0"
    """
    # default for missing coefficient blocks
    default_coeff = "0,0,0,0,0,0"
    pair = preprocess_pair(pair)  # normalize key


    cur = conn.cursor()
    cur.execute("""
        INSERT OR IGNORE INTO binary_params
        (pair, ref, T_min, T_max, eq_num, b0, b1, b2, c_phi, lambda, theta)
        VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
    """, (
        pair,
        ref,
        T_range[0],
        T_range[1],
        eq_num,
        preprocess_data(b0) if b0 is not None else default_coeff,
        preprocess_data(b1) if b1 is not None else default_coeff,
        preprocess_data(b2) if b2 is not None else default_coeff,
        preprocess_data(c_phi) if c_phi is not None else default_coeff,
        preprocess_data(lambda_) if lambda_ is not None else default_coeff,
        preprocess_data(theta) if theta is not None else default_coeff,
    ))
    conn.commit()
    print(f"Inserted/ignored binary parameter for '{pair}'")







if __name__ == "__main__":
    db_path = current_dir / "pypitzer.sqlite"
    conn = sqlite3.connect(db_path)

    
    data = {
    ('Li+,LiCl'):"0,0,0,0,0,0,0,0",
    ('Cl-,LiCl'):"0,0,0,0,0,0,0,0",
    ('K+,LiCl'): "-2.879993582,-1.996147e-3,0,0,0,0.590495613,0.016767927,0",
    ('Na+,LiCl'):"-0.112759482,0.0394693e-3,0,0,0,0,0,0",
}

    for k, v in data.items():
        insert_binary_param(
            conn, 
            pair=k, # test
            ref="Lassin et al., 2015", 
            T_range=[25,300], 
            eq_num=2,
            b0=None, 
            b1=None, 
            b2=None, 
            c_phi=None, 
            lambda_=v, 
            theta=None
        )

